import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Stream;
import java.awt.Point;
import java.io.IOException;

class Main {
  public static void main(String[] args) {
    String fileName = "input.txt";
    try (Stream<String> stream = Files.lines(Paths.get(fileName))) {

      final String[] input = stream.toArray(String[]::new);
      final List<Point> asteroids = parseInput(input);

      Map<Point, List<Point>> environment = new HashMap<Point, List<Point>>();
      final Point station = findStationLocation(asteroids, environment);

      System.out.println("New station at: (" + (int) station.getX() + ", " + (int) station.getY() + ")");
      System.out.println("Number of visible asteroids: " + environment.size());

      final List<Point> sequencing = vaporizeAsteroids(asteroids, environment, station);
      Point p = sequencing.get(199);

      System.out.println("200th explosion at: (" + (int) p.getX() + ", " + (int) p.getY() + ")");

    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  private static List<Point> parseInput(String[] input) {
    List<Point> asteroids = new ArrayList<Point>();
    for (int i = 0; i < input.length; ++i) {
      String[] chars = input[i].split("");
      for (int j = 0; j < chars.length; ++j) {
        if (chars[j].equals("#"))
          asteroids.add(new Point(j, i));
      }
    }
    return asteroids;
  }

  private static Point findStationLocation(List<Point> asteroids, Map<Point, List<Point>> environment) {
    Point station = new Point();
    int maxVisible = 0;
    for (Point ref : asteroids) {
      Map<Point, List<Point>> grouping = new HashMap<Point, List<Point>>();
      for (Point asteroid : asteroids) {
        if (!ref.equals(asteroid)) {
          Point dir = getDirection(ref, asteroid);
          if (grouping.containsKey(dir))
            grouping.get(dir).add(asteroid);
          else
            grouping.put(dir, new ArrayList<Point>(List.of(asteroid)));
        }
      }

      if (maxVisible < grouping.size()) {
        maxVisible = grouping.size();
        station = ref;
        environment.clear();
        environment.putAll(grouping);
      }
    }
    return station;
  }

  private static List<Point> vaporizeAsteroids(List<Point> asteroids, Map<Point, List<Point>> environment, final Point station) {
    Comparator<Point> compareByDistance = (Point p1, Point p2) ->
      Double.compare(p1.distance(station), p2.distance(station));
    Comparator<Point> compareByAngle = (Point p1, Point p2) -> {
      double d1 = (Math.atan2(-1 * p1.getY(), p1.getX()) * -1 + Math.PI / 2) / Math.PI;
      double d2 = (Math.atan2(-1 * p2.getY(), p2.getX()) * -1 + Math.PI / 2) / Math.PI;
      d1 = d1 < 0 ? d1 + 2 : d1;
      d2 = d2 < 0 ? d2 + 2 : d2;
      return Double.compare(d1, d2);
    };

    for (List<Point> pointsInLine : environment.values()) {
      Collections.sort(pointsInLine, compareByDistance);
    }
    List<Point> dirs = new ArrayList<Point>(environment.keySet());
    Collections.sort(dirs, compareByAngle);

    List<Point> sequencing = new ArrayList<Point>(asteroids.size() - 1);
    int hitCnt = 0;
    while (hitCnt < asteroids.size() - 1) {
      for (Point dir : dirs) {
        if (!environment.get(dir).isEmpty()) {
          ++hitCnt;
          sequencing.add(environment.get(dir).remove(0));
        }
      }
    }
    return sequencing;
  }

  private static Point getDirection(Point from, Point to) {
    int dx = (int) (to.getX() - from.getX());
    int dy = (int) (to.getY() - from.getY());
    int adx = Math.abs(dx);
    int ady = Math.abs(dy);
    int gcd = gcdEuclid(Math.max(adx, ady), Math.min(adx, ady));
    return new Point(dx / gcd, dy / gcd);
  }

  private static int gcdEuclid(int n1, int n2) {
    return n2 == 0 ? n1 : gcdEuclid(n2, n1 % n2);
  }
}